# -*- coding: utf-8 -*-
"""Common utility functions.
"""
import copy
import logging

import numpy as np
import typhon as ty
from netCDF4 import Dataset

from konrad import constants


__all__ = [
    'append_description',
    'append_timestep_netcdf',
    'return_if_type',
    'phlev_from_plev',
    'refined_pgrid',
    'get_pressure_grids',
    'ozonesquash',
    'ozone_profile_rcemip',
]

logger = logging.getLogger(__name__)


def append_description(dataset, description=None):
    """Append variable attributes to a given dataset.

    Parameters:
          dataset (xarray.Dataset): Dataset including variables to describe.
          description (dict): Dictionary containing variable descriptions.
            The keys are the variable keys used in the Dataset.
            The values are dictionaries themselves containing attributes
            and their names as keys, e.g.:
                desc = {'T': {'units': 'K', 'standard_name': 'temperature'}}
    """
    if description is None:
        description = constants.variable_description

    for key in dataset.variables:
        if key in description:
            dataset[key].attrs = constants.variable_description[key]


def append_timestep_netcdf(filename, data, timestamp):
    """Append a timestep to an existing netCDF4 file.

    Notes:
        The variables to append to have to exist in the netCDF4 file!

    Parameters:
        filename (str): Path to the netCDF4.
        data (dict{ndarray}): Dict-like object containing the data arrays.
            The key is the variable name and the value is an ``ndarray``, a
            ``pandas.Series`` or an ``xarray.DataArray`` e.g.:
                >>> data = {'T': np.array([290, 295, 300])}

        timestamp (float): Timestamp of values appended.
    """
    # Open netCDF4 file in `append` mode.
    with Dataset(filename, 'a') as nc:
        logging.debug('Append timestep to "{}".'.format(filename))
        t = nc.dimensions['time'].size  # get index to store data.
        nc.variables['time'][t] = timestamp  # append timestamp.

        # Append values for each data variable in ``data``.
        for var in data.data_vars:
            # Append variable if it has a `time` dimension.
            if 'time' in nc[var].dimensions:
                # TODO: Find a cleaner way to handle different data dimensions.
                if 'plev' in nc[var].dimensions:
                    if hasattr(data[var], 'values'):
                        nc.variables[var][t, :] = data[var].values
                    else:
                        nc.variables[var][t, :] = data[var]
                else:
                    if hasattr(data[var], 'values'):
                        nc.variables[var][t] = data[var].values
                    else:
                        nc.variables[var][t] = data[var]


def return_if_type(variable, variablename, expect, default):
    """Return a variable if it matches an expected type.

    Parameters:
          variable: Variable to check.
          variablename (str): Variable name for error message.
          expect (type): Expected variable type.
          default: Default value, if varibale is ``None``.

    Raises:
          TypeError: If variable does not match expected type.
    """
    if variable is None:
        # use a surface with heat capacity as default.
        variable = default
    elif not isinstance(variable, expect):
        raise TypeError(
            'Argument `{name}` has to be of type `{type}`.'.format(
                name=variablename, type=expect.__name__)
        )

    return variable


def phlev_from_plev(fulllevels):
    """Returns the linear interpolated halflevels for given array.

    Parameters:
        fulllevels (ndarray): Pressure at fullevels.

    Returns:
        ndarray: Coordinates at halflevel.

    """
    inter = (fulllevels[1:] + fulllevels[:-1]) / 2
    bottom = fulllevels[0] - 0.5 * (fulllevels[1] - fulllevels[0])
    top = 0
    return np.hstack((bottom, inter, top))


def refined_pgrid(start, stop, num=200, shift=0.5, fixpoint=0.):
    """Create a pressure grid with adjustable distribution in logspace.

    Notes:
          Wrapper for ``typhon.math.squeezable_logspace``.

    Parameters:
        start (float): The starting value of the sequence.
        stop (float): The end value of the sequence.
        num (int): Number of sample to generate (Default is 50).
        shift (float): Factor with which the first stepwidth is
            squeezed in logspace. Has to be between  ``(0, 2)`.
            Values smaller than one compress the gridpoints,
            while values greater than 1 strecht the spacing.
            The default is ``0.5`` (bottom heavy.)
        fixpoint (float): Relative fixpoint for squeezing the grid.
            Has to be between ``[0, 1]``. The  default is ``0`` (bottom).

    Returns:
        ndarray: Pressure grid.
    """
    grid = ty.math.squeezable_logspace(
        start=start, stop=stop, num=num, squeeze=shift, fixpoint=fixpoint
    )

    return grid


def get_pressure_grids(start=1000e2, stop=1, num=200, squeeze=0.5):
    """Create matching pressure levels and half-levels.

    Parameters:
        start (float): Pressure of the lowest half-level (surface) [Pa].
        stop (float): Pressure of the highest half-level (TOA) [Pa].
        num (int): Number of **full** pressure levels.
        squeeze (float): Factor with which the first step width is
            squeezed in logspace. Has to be between ``(0, 2)``.
            Values smaller than one compress the half-levels,
            while values greater than 1 stretch the spacing.
            The default is ``0.5`` (bottom heavy.)

    Returns:
        ndarray, ndarray: Full-level pressure, half-level pressure [Pa].
    """
    phlev = ty.math.squeezable_logspace(start, stop, num + 1, squeeze=squeeze)
    plev = np.exp(0.5 * (np.log(phlev[1:]) + np.log(phlev[:-1])))

    return plev, phlev


def ozonesquash(o3, z, squash):
    """
    Squashes the ozone profile upwards or stretches it downwards, with no
        change to the shape of the profile above the ozone concentration maximum
    Parameters:
        o3 (ndarray): initial ozone profile
        z (ndarray): corresponding height values
        squash: float, with 1 being no squash,
            numbers < 1 squashing the profile towards the maximum,
            numbers > 1, stretching the profile downwards
    Returns:
        ndarray: new ozone profile
    """
    i_max_o3 = np.argmax(o3)

    sqz = (z - z[i_max_o3])*squash + z[i_max_o3]
    new_o3 = copy.copy(o3)
    new_o3[:i_max_o3] = np.interp(z[:i_max_o3], sqz, o3)
    return new_o3


def ozone_profile_rcemip(plev, g1=3.6478, g2=0.83209, g3=11.3515):
    """Compute the ozone volumetric mixing ratio from pressure.

    .. math::
        O_3 = g_1 \cdot p^{g_2} e^\frac{-p}{g_3}

    Parameters:
        plev (ndarray): Atmospheric pressure [Pa].
        g1, g2, g3 (float): Fitting parameters for gamma distribution
            according to Wing et al. (2017).

    Returns:
          ndarray: Ozone profile [VMR].

    Reference:
        Wing et al., 2017, Radiative-Convective Equilibrium Model
        Intercomparison Project

    """
    p = plev / 100
    return g1 * p**g2 * np.exp(-p / g3) * 1e-6



"""Define an interface for the RRTMG radiation scheme (through CliMT). """
import numpy as np
import xarray as xr
from sympl import DataArray
from typhon.physics import vmr2specific_humidity

from .radiation import Radiation


__all__ = [
    'RRTMG',
]


class RRTMG(Radiation):
    """RRTMG radiation scheme using the CliMT python wrapper."""

    def __init__(self, *args, solar_constant=510, **kwargs):
        super().__init__(*args, **kwargs)
        self.state_lw = None
        self.state_sw = None

        self.solar_constant = solar_constant

    def update_radiative_state(self, atmosphere, state0, sw=True):
        """ Update CliMT formatted atmospheric state using parameters from our
        model.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
            state0 (dictionary): atmospheric state in the format for climt
            sw (bool): Toggle between shortwave and longwave calculations.

        Returns:
            dictionary: updated state
        """
        plev = atmosphere['plev'].values
        phlev = atmosphere['phlev'].values
        numlevels = len(plev)
        temperature = atmosphere.surface.temperature.data[0]
        albedo = float(atmosphere.surface.albedo.data)
        o2fraction = 0.21
        zenith = np.deg2rad(self.current_solar_angle)

        state0['mid_levels'] = DataArray(
            np.arange(0, numlevels),
            dims=('mid_levels',),
            attrs={'label': 'mid_levels'})

        state0['interface_levels'] = DataArray(
            np.arange(0, numlevels+1),
            dims=('interface_levels',),
            attrs={'label': 'interface_levels'})

        state0['air_pressure'] = DataArray(
            plev,
            dims=('mid_levels',),
            attrs={'units': 'Pa'})

        state0['air_pressure_on_interface_levels'] = DataArray(
            phlev,
            dims=('interface_levels',),
            attrs={'units': 'Pa'})

        state0['air_temperature'] = DataArray(
            atmosphere['T'][0, :].data,
            dims=('mid_levels',),
            attrs={'units': 'degK'})

        vmr_h2o = atmosphere['H2O'][0, :].data
        specific_humidity = vmr2specific_humidity(vmr_h2o)
        state0['specific_humidity'] = DataArray(
            specific_humidity,
            dims=('mid_levels',),
            attrs={'units': 'g/g'})

        state0['mole_fraction_of_oxygen_in_air'] = DataArray(
            o2fraction * np.ones(numlevels,),
            dims=('mid_levels',),
            attrs={'units': 'mole/mole'})

        # Set trace gas concentrations
        for gasname, gas in [('mole_fraction_of_methane_in_air', 'CH4'),
                             ('mole_fraction_of_carbon_dioxide_in_air', 'CO2'),
                             ('mole_fraction_of_nitrous_oxide_in_air', 'N2O'),
                             ('mole_fraction_of_ozone_in_air', 'O3')]:
            state0[gasname] = DataArray(
                atmosphere[gas][0, :].data,
                dims=('mid_levels',),
                attrs={'units': 'mole/mole'})

        for gasname in ['mole_fraction_of_carbon_tetrachloride_in_air',
                        'mole_fraction_of_cfc22_in_air',
                        'mole_fraction_of_cfc12_in_air',
                        'mole_fraction_of_cfc11_in_air']:
            state0[gasname] = DataArray(
                np.zeros(numlevels,),
                dims=('mid_levels',),
                attrs={'units': 'mole/mole'})

        # Set cloud quantities to zero.
        for quant in ['cloud_ice_particle_size',
                      'mass_content_of_cloud_liquid_water_in_atmosphere_layer',
                      'mass_content_of_cloud_ice_in_atmosphere_layer',
                      'cloud_water_droplet_radius',
                      'cloud_area_fraction_in_atmosphere_layer']:
            units = state0[quant].units
            state0[quant] = DataArray(
                np.zeros(numlevels,),
                dims=('mid_levels',),
                attrs={'units': units})

        if not sw:  # Longwave specific changes
            num_lw_bands = len(state0['num_longwave_bands'])

            state0['longwave_optical_thickness_due_to_cloud'] = DataArray(
                np.zeros((1, num_lw_bands, 1, numlevels)),
                dims=('longitude', 'num_longwave_bands',
                      'latitude', 'mid_levels'),
                attrs={'units': 'dimensionless'})

            state0['longwave_optical_thickness_due_to_aerosol'] = DataArray(
                np.zeros((1, 1, numlevels, num_lw_bands)),
                dims=('longitude', 'latitude',
                      'mid_levels', 'num_longwave_bands'),
                attrs={'units': 'dimensionless'})

        if sw:  # Shortwave specific changes
            num_sw_bands = len(state0['num_shortwave_bands'])

            for quant in ['cloud_forward_scattering_fraction',
                          'cloud_asymmetry_parameter',
                          'shortwave_optical_thickness_due_to_cloud',
                          'single_scattering_albedo_due_to_cloud']:
                state0[quant] = DataArray(
                    np.zeros((1, num_sw_bands, 1, numlevels)),
                    dims=('longitude', 'num_shortwave_bands',
                          'latitude', 'mid_levels'),
                    attrs={'units': 'dimensionless'})

            num_aerosols = len(state0['num_ecmwf_aerosols'])
            state0['aerosol_optical_depth_at_55_micron'] = DataArray(
                np.zeros((1, 1, numlevels, num_aerosols)),
                dims=('longitude', 'latitude',
                      'mid_levels', 'num_ecmwf_aerosols'),
                attrs={'units': 'dimensionless'})

            for quant in ['shortwave_optical_thickness_due_to_aerosol',
                          'single_scattering_albedo_due_to_aerosol',
                          'aerosol_asymmetry_parameter']:
                state0[quant] = DataArray(
                    np.zeros((1, 1, numlevels, num_sw_bands)),
                    dims=('longitude', 'latitude',
                          'mid_levels', 'num_shortwave_bands'),
                    attrs={'units': 'dimensionless'})

        # TODO: Should all the aerosol values be zero?!

        # Surface quantities
        state0['surface_temperature'] = DataArray(
            np.array([[temperature]]),
            dims={'longitude', 'latitude'},
            attrs={'units': 'degK'})

        if sw:  # surface properties required only for shortwave
            for surface_albedo in ['surface_albedo_for_diffuse_near_infrared',
                                   'surface_albedo_for_direct_near_infrared',
                                   'surface_albedo_for_diffuse_shortwave',
                                   'surface_albedo_for_direct_shortwave']:
                state0[surface_albedo] = DataArray(
                    np.array([[albedo]]),
                    dims={'longitude', 'latitude'},
                    attrs={'units': 'dimensionless'})

            # Sun
            state0['zenith_angle'] = DataArray(
                np.array([[zenith]]),
                dims={'longitude', 'latitude'},
                attrs={'units': 'radians'})

        return state0

    def radiative_fluxes(self, atmosphere):
        """Returns shortwave and longwave fluxes and heating rates.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): atmosphere model

        Returns:
            tuple: containing two dictionaries, one of air temperature
            values and the other of fluxes and heating rates
        """
        if self.state_lw is None or self.state_sw is None:
            import climt
            climt.set_constant('stellar_irradiance',
                               value=self.solar_constant,
                               units='W m^-2')
            self.rad_lw = climt.RRTMGLongwave()
            self.rad_sw = climt.RRTMGShortwave(ignore_day_of_year=True)
            self.state_lw = climt.get_default_state([self.rad_lw])
            self.state_sw = climt.get_default_state([self.rad_sw])

        self.update_radiative_state(atmosphere, self.state_lw, sw=False)
        self.update_radiative_state(atmosphere, self.state_sw, sw=True)

        lw_fluxes = self.rad_lw(self.state_lw)
        sw_fluxes = self.rad_sw(self.state_sw)

        return lw_fluxes, sw_fluxes

    def calc_radiation(self, atmosphere):
        """Returns the shortwave, longwave and net heatingrates.
        Converts output from radiative_fluxes to be in the format required for
        our model.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.

        Returns:
            xarray.Dataset: Dataset containing for the simulated heating rates.
                The keys are 'sw_htngrt', 'lw_htngrt' and 'net_htngrt'.
        """
        lw_dT_fluxes, sw_dT_fluxes = self.radiative_fluxes(atmosphere)
        lw_fluxes = lw_dT_fluxes[1]
        sw_fluxes = sw_dT_fluxes[1]

        ret = xr.Dataset({
            # General atmospheric properties.
            'z': atmosphere['z'],

            # Longwave fluxes and heatingrates.
            'lw_htngrt': (
                ['time', 'plev'],
                lw_fluxes['longwave_heating_rate'][0].data),
            'lw_htngrt_clr': (
                ['time', 'plev'],
                lw_fluxes['longwave_heating_rate_assuming_clear_sky'][0].data),
            'lw_flxu': (
                ['time', 'phlev'],
                lw_fluxes['upwelling_longwave_flux_in_air'][0].data),
            'lw_flxd': (
                ['time', 'phlev'],
                lw_fluxes['downwelling_longwave_flux_in_air'][0].data),
            'lw_flxu_clr': (
                ['time', 'phlev'],
                lw_fluxes['upwelling_longwave_flux_in_air_assuming_clear_sky'][0].data),
            'lw_flxd_clr': (
                ['time', 'phlev'],
                lw_fluxes['downwelling_longwave_flux_in_air_assuming_clear_sky'][0].data),

            # Shortwave fluxes and heatingrates.
            'sw_htngrt': (
                ['time', 'plev'],
                sw_fluxes['shortwave_heating_rate'][0].data),
            'sw_htngrt_clr': (
                ['time', 'plev'],
                sw_fluxes['shortwave_heating_rate_assuming_clear_sky'][0].data),
            'sw_flxu': (
                ['time', 'phlev'],
                sw_fluxes['upwelling_shortwave_flux_in_air'][0].data),
            'sw_flxd': (
                ['time', 'phlev'],
                sw_fluxes['downwelling_shortwave_flux_in_air'][0].data),
            'sw_flxu_clr': (
                ['time', 'phlev'],
                sw_fluxes['upwelling_shortwave_flux_in_air_assuming_clear_sky'][0].data),
            'sw_flxd_clr': (
                ['time', 'phlev'],
                sw_fluxes['downwelling_shortwave_flux_in_air_assuming_clear_sky'][0].data),
        },
            coords={
                'time': [0],
                'phlev': atmosphere['phlev'].values,
                'plev': atmosphere['plev'].values,
            }
        )

        return ret



# -*- coding: utf-8 -*-
"""This module contains classes handling different convection schemes."""
import abc

import numpy as np
import typhon
from scipy.interpolate import interp1d

from konrad import constants
from konrad.surface import SurfaceFixedTemperature


__all__ = [
    'Convection',
    'NonConvective',
    'HardAdjustment',
    'RelaxedAdjustment',
]


def energy_difference(T_2, T_1, sst_2, sst_1, dp, eff_Cp_s):
    """
    Calculate the energy difference between two atmospheric profiles (2 - 1).

    Parameters:
        T_2: atmospheric temperature profile (2)
        T_1: atmospheric temperature profile (1)
        sst_2: surface temperature (2)
        sst_1: surface temperature (1)
        dp: pressure thicknesses of levels,
            must be the same for both atmospheric profiles
        eff_Cp_s: effective heat capacity of surface
    """
    Cp = constants.isobaric_mass_heat_capacity
    g = constants.g

    dT = T_2 - T_1  # convective temperature change of atmosphere
    dT_s = sst_2 - sst_1  # of surface
    termdiff = - np.sum(Cp/g * dT * dp) + eff_Cp_s * dT_s

    return termdiff


class Convection(metaclass=abc.ABCMeta):
    """Base class to define abstract methods for convection schemes."""
    @abc.abstractmethod
    def stabilize(self, atmosphere, lapse, timestep):
        """Stabilize the temperature profile by redistributing energy.

        Parameters:
              atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
              lapse (ndarray): Temperature lapse rate [K/day].
              timestep (float): Timestep width [day].
        """


class NonConvective(Convection):
    """Do not apply convection."""
    def stabilize(self, *args, **kwargs):
        pass


class HardAdjustment(Convection):
    """Instantaneous adjustment of temperature profiles"""
    def stabilize(self, atmosphere, lapse, timestep):

        # Find convectively adjusted temperature profile.
        T_new, T_s_new = self.convective_adjustment(
            p=atmosphere['plev'].values,
            phlev=atmosphere['phlev'].values,
            T_rad=atmosphere['T'].values[0, :],
            lapse=lapse,
            surface=atmosphere.surface,
            timestep=timestep,
        )
        # Update atmospheric temperatures as well as surface temperature.
        atmosphere['T'].values[0, :] = T_new
        atmosphere.surface.temperature.values[0] = T_s_new

    def convective_adjustment(self, p, phlev, T_rad, lapse, surface,
                              timestep=0.1):
        """
        Find the energy-conserving temperature profile using a iterative
        procedure with test profiles. Update the atmospheric temperature
        profile to this one.

        Parameters:
            p (ndarray): pressure levels
            phlev (ndarray): half pressure levels
            T_rad (ndarray): old atmospheric temperature profile
            lapse (konrad.lapserate): lapse rate in K/km
            surface (konrad.surface):
                surface associated with old temperature profile
            timestep (float): only required for slow convection
        """
        # The threshold is scaled with the effective heat capacity of the
        # surface. Otherwise very thick surfaces may never reach the target.
        near_zero = float(surface.heat_capacity / 1e13)

        # Interpolate density and lapse rate on pressure half-levels.
        density1 = typhon.physics.density(p, T_rad)
        density = interp1d(p, density1, fill_value='extrapolate')(phlev[:-1])

        g = constants.earth_standard_gravity
        lp = -lapse / (g * density)
        lp = lp.data

        # find energy difference if there is no change to surface temp due to
        # convective adjustment. in this case the new profile should be
        # associated with an increase in energy in the atmosphere.
        surfaceTpos = surface.temperature.data
        T_con, diffpos = self.test_profile(T_rad, p, phlev, surface,
                                           surfaceTpos, lp,
                                           timestep=timestep)
        
        # this is the temperature profile required if we have a set-up with a
        # fixed surface temperature, then the energy does not matter.
        if isinstance(surface, SurfaceFixedTemperature):
            return T_con, surface.temperature
        # for other cases, if we find a decrease or approx no change in energy,
        # the atmosphere is not being warmed by the convection,
        # as it is not unstable to convection, so no adjustment is applied
        if diffpos < near_zero:
            return T_con, surface.temperature

        # if the atmosphere is unstable to convection, a fixed surface temp
        # produces an increase in energy (as convection warms the atmosphere).
        # this surface temperature is an upper bound to the energy-conserving
        # surface temperature.
        # taking the surface temperature as the coldest temperature in the
        # radiative profile gives us a lower bound.
        surfaceTneg = np.array([np.min(T_rad)])
        eff_Cp_s = surface.heat_capacity
        diffneg = eff_Cp_s.data * (surfaceTneg - surface.temperature.data)
        # good guess for energy-conserving profile (unlikely!)
        if np.abs(diffneg) < near_zero:
            return T_con, surfaceTneg

        # NOTE (lkluft): Dirty workaround to always initialize `surfaceT`.
        # I encountered situations where the while-loop did not run a single
        # iteration and therefore the return-statement failed.
        surfaceT = (surfaceTneg + (surfaceTpos - surfaceTneg)
                    * (-diffneg) / (-diffneg + diffpos))

        # Now we have a upper and lower bound for the surface temperature of
        # the energy conserving profile. Iterate to get closer to the energy-
        # conserving temperature profile.
        counter = 0
        while diffpos >= near_zero and -diffneg >= near_zero:
            surfaceT = (surfaceTneg + (surfaceTpos - surfaceTneg)
                        * (-diffneg) / (-diffneg + diffpos))
            T_con, diff = self.test_profile(T_rad, p, phlev, surface, surfaceT,
                                            lp, timestep=timestep)
            if diff > 0:
                diffpos = diff
                surfaceTpos = surfaceT
            else:
                diffneg = diff
                surfaceTneg = surfaceT

            # to avoid getting stuck in a loop if something weird is going on
            counter += 1
            if counter == 100:
                raise ValueError(
                    "No energy conserving convective profile can be found"
                )

        # save new temperature profile
        return T_con, surfaceT

    def test_profile(self, T_rad, p, phlev, surface, surfaceT, lp,
                     timestep=0.1):
        """
        Assuming a particular surface temperature (surfaceT), create a new
        profile, following the specified lapse rate (lp) for the region where
        the convectively adjusted atmosphere is warmer than the radiative one.

        Parameters:
            T_rad (ndarray): old atmospheric temperature profile
            p (ndarray): pressure levels
            phlev (ndarray): half pressure levels
            surface (konrad.surface):
                surface associated with old temperature profile 
            surfaceT (float): surface temperature of the new profile
            lp (ndarray): lapse rate in K/Pa
            timestep (float): not required in this case

        Returns:
            ndarray: new atmospheric temperature profile
            float: energy difference between the new profile and the old one
        """
        # dp, thicknesses of atmosphere layers, for energy calculation
        dp = np.diff(phlev)
        # for lapse rate integral
        dp_lapse = np.hstack((np.array([p[0] - phlev[0]]), np.diff(p)))
        T_con = surfaceT - np.cumsum(dp_lapse * lp)
        if np.any(T_con > T_rad):
            contop = np.max(np.where(T_con > T_rad))
            T_con[contop+1:] = T_rad[contop+1:]
        else:
            T_con = T_rad
            contop = 0

        # If run with a fixed surface temperature, always return the
        # convective profile starting from the current surface temperature.
        if isinstance(surface, SurfaceFixedTemperature):
            return T_con, 0.

        eff_Cp_s = surface.heat_capacity

        diff = energy_difference(T_con, T_rad, surfaceT, surface.temperature,
                                 dp, eff_Cp_s)

        return T_con, float(diff)


class RelaxedAdjustment(HardAdjustment):
    """Adjustment with relaxed convection in upper atmosphere.

    This convection scheme allows for a transition regime between a
    convectively driven troposphere and the radiatively balanced stratosphere.
    """
    def __init__(self, tau=None):
        """
        Parameters:
            tau (ndarray): Array of convective timescale values [days]
        """        
        self.convective_tau = tau

    def get_convective_tau(self, p):
        """Return a convective timescale profile.

        Parameters:
            p (ndarray): Pressure levels [Pa].

        Returns:
            ndarray: Convective timescale profile [days].
        """
        if self.convective_tau is not None:
            return self.convective_tau

        tau0 = 1/24 # 1 hour
        tau = tau0*np.exp(p[0] / p)
        
        return tau

    def test_profile(self, T_rad, p, phlev, surface, surfaceT, lp,
                     timestep=0.1):
        """
        Assuming a particular surface temperature (surfaceT), create a new
        profile, using the convective timescale and specified lapse rate (lp).

        Parameters:
            T_rad (ndarray): old atmospheric temperature profile
            p (ndarray): pressure levels
            phlev (ndarray): half pressure levels
            surface (konrad.surface):
                surface associated with old temperature profile 
            surfaceT (float): surface temperature of the new profile
            lp (ndarray): lapse rate in K/Pa
            timestep (float): not required in this case

        Returns:
            ndarray: new atmospheric temperature profile
            float: energy difference between the new profile and the old one
        """
        dp = np.diff(phlev)
        dp_lapse = np.hstack((np.array([p[0] - phlev[0]]), np.diff(p)))

        tau = self.get_convective_tau(p)
            
        tf = 1 - np.exp(-timestep / tau)
        T_con = T_rad * (1 - tf) + tf * (surfaceT - np.cumsum(dp_lapse * lp))
        
        # If run with a fixed surface temperature, always return the
        # convective profile starting from the current surface temperature.
        if isinstance(surface, SurfaceFixedTemperature):
            return T_con, 0.
        
        eff_Cp_s = surface.heat_capacity

        diff = energy_difference(T_con, T_rad, surfaceT, surface.temperature,
                                 dp, eff_Cp_s)
        return T_con, float(diff)





"""Define an interface for the PSRAD radiation scheme. """
import numpy as np
import xarray as xr

from .radiation import Radiation


__all__ = [
    'PSRAD',
]


class PSRAD(Radiation):
    """Radiation model using the ICON PSRAD radiation scheme.

    The PSRAD class relies on external radiative-transfer code. Currently this
    functionality is provided by the PSRAD radiation scheme. You need a
    compiled version of this code in order to install and run ``konrad``.

    A stable version is accessible through the internal subversion repository:

        $ svn co https://arts.mi.uni-hamburg.de/svn/internal/psrad/trunk psrad

    Follow the instructions given in the repository to compile PSRAD on your
    machine. A part of the installation process is to set some environment
    variables. Thos are also needed in order to run ``konrad``:

        $ source config/psrad_env.bash

    """
    @staticmethod
    def _extract_psrad_args(atmosphere):
        """Returns tuple of mixing ratios to use with psrad.

        Paramteres:
            atmosphere (dict or pandas.DataFrame): Atmospheric atmosphere.

        Returns:
            tuple(ndarray): ndarrays in the order and unit to use with `psrad`:
                Z, P, T, x_vmr, ...
        """
        z = atmosphere['z'].values
        p = atmosphere['plev'].values / 100
        T = atmosphere['T'].values

        ret = [z, p, T]  # Z, P, T

        # Keep order as it is expected by PSRAD.
        required_gases = ['H2O', 'O3', 'CO2', 'N2O', 'CO', 'CH4']

        for gas in required_gases:
            if gas in atmosphere:
                ret.append(atmosphere[gas].values * 1e6)  # Append gas in ppm.
            else:
                ret.append(np.zeros(np.size(p)))

        return tuple(ret)

    def calc_radiation(self, atmosphere):
        """Returns the shortwave, longwave and net heatingrates.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.

        Returns:
            xarray.Dataset: Dataset containing for the simulated heating rates.
                The keys are 'sw_htngrt', 'lw_htngrt' and 'net_htngrt'.
        """
        from . import _psrad
        self.psrad = _psrad

        dmy_indices = np.asarray([0, 0, 0, 0])
        ic = dmy_indices.astype("int32") + 1
        c_lwc = np.asarray([0., 0., 0., 0.])
        c_iwc = np.asarray([0., 0., 0., 0.])
        c_frc = np.asarray([0., 0., 0., 0.])

        nlev = atmosphere['plev'].size

        # Extract surface properties.
        P_sfc = atmosphere.surface.pressure.values / 100
        T_sfc = atmosphere.surface.temperature.values[0]
        albedo = atmosphere.surface.albedo.values

        # Use the **current** solar angle as zenith angle for the simulation.
        zenith = self.current_solar_angle

        self.psrad.setup_single(nlev, len(ic), ic, c_lwc, c_iwc, c_frc, zenith,
                                albedo, P_sfc, T_sfc,
                                *self._extract_psrad_args(atmosphere))

        self.psrad.advance_lrtm()  # Longwave simulations.
        (lw_hr, lw_hr_clr, lw_flxd,
         lw_flxd_clr, lw_flxu, lw_flxu_clr) = self.psrad.get_lw_fluxes()

        self.psrad.advance_srtm()  # Shortwave simulations.

        (sw_hr, sw_hr_clr, sw_flxd,
         sw_flxd_clr, sw_flxu, sw_flxu_clr,
         vis_frc, par_dn, nir_dff, vis_diff,
         par_diff) = self.psrad.get_sw_fluxes()

        ret = xr.Dataset({
            # General atmospheric properties.
            'z': atmosphere['z'],
            # Longwave fluxes and heatingrates.
            'lw_htngrt': (['time', 'plev'], lw_hr[:, :]),
            'lw_htngrt_clr': (['time', 'plev'], lw_hr_clr[:, :]),
            'lw_flxu': (['time', 'phlev'], lw_flxu[:, :]),
            'lw_flxd': (['time', 'phlev'], lw_flxd[:, :]),
            'lw_flxu_clr': (['time', 'phlev'], lw_flxu_clr[:, :]),
            'lw_flxd_clr': (['time', 'phlev'], lw_flxd_clr[:, :]),
            # Shortwave fluxes and heatingrates.
            # Note: The shortwave fluxes and heatingrates calculated by PSRAD
            # are **inverted**. Therefore, they are flipped to make the input
            # and output of this function consistent.
            'sw_htngrt': (['time', 'plev'], sw_hr[:, ::-1]),
            'sw_htngrt_clr': (['time', 'plev'], sw_hr_clr[:, ::-1]),
            'sw_flxu': (['time', 'phlev'], sw_flxu[:, ::-1]),
            'sw_flxd': (['time', 'phlev'], sw_flxd[:, ::-1]),
            'sw_flxu_clr': (['time', 'phlev'], sw_flxu_clr[:, ::-1]),
            'sw_flxd_clr': (['time', 'phlev'], sw_flxd_clr[:, ::-1]),
        },
            coords={
                'time': [0],
                'plev': atmosphere['plev'].values,
                'phlev': atmosphere['phlev'].values,
            }
        )

        return ret



"""Module PSRAD: ICON radiation library interface.
The functions in this module provide interface to the PSRAD radiation library used in the ICON model. The simplest usage consists of:

* calling "setup" with all relevant input data;
* calling "advance_lrtm" and/or "advance_srtm" to compute the respective fluxes;
* calling "get_lw_fluxes" and/or "get_sw_fluxes" to obtain the results.

There are, in addition, query functions for variables internal to the module.

The package must be able to find "libpsrad.so.1" and its associated "*.nc" data files, all of which should reside in the same path from which python is first invoked.

This interface uses the Numpy module, and all arrays passed as arguments to or returned by functions are Numpy arrays. Integer and floating-point arrays must be of types "int32" and "float64" respectively, othewise the module will raise and exception.
"""
from ctypes import cdll as __cdll
from ctypes import POINTER as __POINTER
from ctypes import c_int, c_double
solib = __cdll.LoadLibrary('./libpsrad.so.1')
_c_int_p = __POINTER(c_int)
_c_double_p = __POINTER(c_double)

def __get_double_p(a):
  from numpy import dtype
  from ctypes import POINTER
  if (a.dtype != dtype('float64')):
    print("parameter is ", a.dtype.str, "(expected float64)")
    raise()
  return a.ctypes.data_as(POINTER(c_double))

def __get_int_p(a):
  from numpy import dtype
  from ctypes import POINTER
  if (a.dtype != dtype('int32')):
    print("parameter is ", a.dtype.str, "(expected int32)")
    raise()
  return a.ctypes.data_as(POINTER(c_int))

solib.psrad_get_kbdim.restype = c_int
def get_kbdim():
  "Returns (int32) the number of columns, as specified in the call to 'setup'."
  return solib.psrad_get_kbdim()

solib.psrad_get_klev.restype = c_int
def get_klev():
  "Returns (int32) the number of layers, as specified in the call to 'setup'."
  return solib.psrad_get_klev()

solib.psrad_get_nlayers.restype = c_int
def get_nlayers():
  "Returns (int32) the number of cloud layers, as specified in the call to 'setup'"
  return solib.psrad_get_nlayers()

def get_strategy_lw():
  "Returns (int32) the LW sampling strategy. Currently hard-coded."
  return 1

def get_strategy_sw():
  "Returns (int32) the SW sampling strategy. Currently hard-coded."
  return 1

solib.psrad_get_num_gpts_lw.restype = c_int
def get_num_gpts_lw():
  "Returns (int32) the total number of g-points in the long-wave range."
  return solib.psrad_get_ngptlw()

solib.psrad_get_num_gpts_sw.restype = c_int
def get_num_gpts_sw():
  "Returns (int32) the total number of g-points in the short-wave range."
  return solib.psrad_get_ngptsw()

solib.psrad_get_nbndlw.restype = c_int
def get_nbandlw():
  "Returns (int32) the number of long-wave bands. Currently hard-coded."
  return solib.psrad_get_nbndlw()

solib.psrad_get_nbndsw.restype = c_int
def get_nbandsw():
  "Returns (int32) the number of short-wave bands. Currently hard-coded."
  return solib.psrad_get_nbndsw()

solib.psrad_get_prmu0.restype = c_double
def get_prmu0():
  "Returns (float64) cos(zenith) for the value specified in the call to 'setup'."
  return solib.psrad_get_prmu0()


solib.psrad_standalone_setup_single.argtypes = [c_int, \
c_int, _c_int_p, _c_double_p, _c_double_p, _c_double_p, c_double, c_double, \
c_double, c_double, _c_double_p, _c_double_p, _c_double_p, _c_double_p, \
_c_double_p, _c_double_p, _c_double_p,_c_double_p]
def setup_single(klev, nlayers, icldlyr_index, cldlyr_lwc, cldlyr_iwc, \
cldlyr_frc, zenith, albedo, pp_sfc, tk_sfc, hgt_fl_vr, pp_fl_vr, tk_fl_vr, \
xm_h2o_vr, xm_o3_vr, xm_co2_vr, xm_n2o_vr, xm_co_vr, xm_ch4_vr):
  """Setup radiation calculation. Input data is as follows:
* klev (int): number of layers in the column
* nlayers (int): number of layers with clouds
* icldlyr_index (int[nlayers]): indexes (in [1:klev]) of layers with clouds. NOTICE THESE INDICES ARE IN FORTRAN 1-OFFSET CONVENTION.
* cldlyr_lwc (float64[nlayers]): Liquid water content in cloud layer. [???]
* cldlyr_iwc (float64[nlayers]): Ice water content in cloud layer. [???]
* cldlyr_frc (float64[nlayers]): Cloud fraction in cloud layer (in [0,1]).
* zenith (float64): Solar zenith angle [deg]
* albedo (float64): constant, band independent albedo (in [0,1]).
* pp_sfc (float64): Surface pressure [Pa]
* tk_sfc (float64): Surface temperature [K]
* hgt_fl_vr (float64[klev]): Full level heights [mb]
* pp_fl_vr (float64[klev]): Full level pressure [Pa]
* tk_fl_vr (float64[klev]): Full level temperature [K]
* xm_*_vr (float64[klev]): Volume mixing rations for chemical species *. [???]
The last variables are indexed from the ground up, i.e. pp_fl_vr[0] is the pressure at the surface and pp_fl_vr[klev-1] is the TOA pressure."""
  l_lw = c_int(1)
  l_sw = c_int(1)
  klev = c_int(klev)
  nlayers = c_int(len(icldlyr_index))

  C_icldlyr_index = __get_int_p(icldlyr_index)
  C_cldlyr_lwc = __get_double_p(cldlyr_lwc) 
  C_cldlyr_iwc = __get_double_p(cldlyr_iwc)
  C_cldlyr_frc = __get_double_p(cldlyr_frc)
  C_pp_sfc = c_double(pp_sfc)
  C_tk_sfc = c_double(tk_sfc)
  C_hgt_fl_vr = __get_double_p(hgt_fl_vr)
  C_pp_fl_vr = __get_double_p(pp_fl_vr)
  C_tk_fl_vr = __get_double_p(tk_fl_vr)
  C_xm_h2o_vr = __get_double_p(xm_h2o_vr)
  C_xm_o3_vr = __get_double_p(xm_o3_vr)
  C_xm_co2_vr = __get_double_p(xm_co2_vr)
  C_xm_n2o_vr = __get_double_p(xm_n2o_vr)
  C_xm_co_vr = __get_double_p(xm_co_vr)
  C_xm_ch4_vr = __get_double_p(xm_ch4_vr)
  C_zenith = c_double(zenith)
  C_albedo = c_double(albedo)
  
  solib.psrad_standalone_setup_single(klev, nlayers, C_icldlyr_index, C_cldlyr_lwc, C_cldlyr_iwc, C_cldlyr_frc, C_zenith, C_albedo, C_pp_sfc, C_tk_sfc, C_hgt_fl_vr, C_pp_fl_vr, C_tk_fl_vr, C_xm_h2o_vr, C_xm_o3_vr, C_xm_co2_vr, C_xm_n2o_vr, C_xm_co_vr, C_xm_ch4_vr)

solib.psrad_standalone_setup_multi.argtypes = [c_int, c_int, _c_int_p, \
_c_double_p, _c_double_p, _c_double_p, c_double, c_double, \
_c_double_p, _c_double_p, _c_double_p, _c_double_p, _c_double_p, _c_double_p, \
_c_double_p, _c_double_p, _c_double_p,_c_double_p]
def setup_multi(kbdim, klev, icldlyr, cldlyr_lwc, cldlyr_iwc, \
cldlyr_frc, zenith, albedo, pp_sfc, tk_sfc, hgt_fl_vr, pp_fl_vr, tk_fl_vr, \
xm_h2o_vr, xm_o3_vr, xm_co2_vr, xm_n2o_vr, xm_co_vr, xm_ch4_vr):
  """Setup radiation calculation. Input data is as follows:
* kbdim (int): number of columns
* klev (int): number of layers per column
* icldlyr (int[kbdim,klev]): flag for cells with clouds.
* cldlyr_lwc (float64[kbdim,klev]): Liquid water content in cloud layer. [???]
* cldlyr_iwc (float64[kbdim,klev]): Ice water content in cloud layer. [???]
* cldlyr_frc (float64[kbdim,klev]): Cloud fraction in cloud layer (in [0,1]).
* zenith (float64): Solar zenith angle [deg]
* albedo (float64): constant, band independent albedo (in [0,1]).
* pp_sfc (float64[kbdim]): Surface pressure [Pa]
* tk_sfc (float64[kbdim]): Surface temperature [K]
* hgt_fl_vr (float64[kbdim,klev]): Full level heights [mb]
* pp_fl_vr (float64[kbdim,klev]): Full level pressure [Pa]
* tk_fl_vr (float64[kbdim,klev]): Full level temperature [K]
* xm_*_vr (float64[kbdim,klev]): Volume mixing rations for chemical species *. [???]
The last variables are indexed from the ground up, i.e. pp_fl_vr[:,0] is the pressure at the surface and pp_fl_vr[:,klev-1] is the TOA pressure."""
  l_lw = c_int(1)
  l_sw = c_int(1)
  kbdim = c_int(kbdim)
  klev = c_int(klev)

  C_icldlyr = __get_int_p(icldlyr)
  C_cldlyr_lwc = __get_double_p(cldlyr_lwc) 
  C_cldlyr_iwc = __get_double_p(cldlyr_iwc)
  C_cldlyr_frc = __get_double_p(cldlyr_frc)
  C_pp_sfc = __get_double_p(pp_sfc)
  C_tk_sfc = __get_double_p(tk_sfc)
  C_hgt_fl_vr = __get_double_p(hgt_fl_vr)
  C_pp_fl_vr = __get_double_p(pp_fl_vr)
  C_tk_fl_vr = __get_double_p(tk_fl_vr)
  C_xm_h2o_vr = __get_double_p(xm_h2o_vr)
  C_xm_o3_vr = __get_double_p(xm_o3_vr)
  C_xm_co2_vr = __get_double_p(xm_co2_vr)
  C_xm_n2o_vr = __get_double_p(xm_n2o_vr)
  C_xm_co_vr = __get_double_p(xm_co_vr)
  C_xm_ch4_vr = __get_double_p(xm_ch4_vr)
  C_zenith = c_double(zenith)
  C_albedo = c_double(albedo)
  
  solib.psrad_standalone_setup_multi(kbdim, klev, C_icldlyr, C_cldlyr_lwc, C_cldlyr_iwc, C_cldlyr_frc, C_zenith, C_albedo, C_pp_sfc, C_tk_sfc, C_hgt_fl_vr, C_pp_fl_vr, C_tk_fl_vr, C_xm_h2o_vr, C_xm_o3_vr, C_xm_co2_vr, C_xm_n2o_vr, C_xm_co_vr, C_xm_ch4_vr)

def advance_lrtm():
  "Execute long wave radiation routines. Results are then available through the 'get_lw_fluxes' function."
  solib.psrad_advance_lrtm()
def advance_srtm():
  "Execute short wave radiation routines. Results are then available through the 'get_sw_fluxes' function."
  solib.psrad_advance_srtm()

def get_lw_fluxes():
  """Returns the following arrays with the results of the long wave routines:
* htngrt(float64[klev]): Heating rate of total sky
* htngrt_clr(float64[klev]): Heating rate of clear sky
For these variables, position [0] is the value at the surface layer and position [klev-1] is at the TOA layer. NOTICE THIS IS THE OPPOSITE OF THE OTHER ROUTINE
* flxd_lw_vr(float64[klev+1]): Downward flux of total sky
* flxd_lw_clr_vr(float64[klev+1]): Downward flux of clear sky
* flxu_lw_vr(float64[klev+1]): Upward flux of total sky. 
* flxu_lw_clr_vr(float64[klev+1]): Upward flux of clear sky
For these variables, position [i] is the value at the bottom of layer [i], with 0 being the surface and [klev-1] being the last layer. Position [klev] is the TOA value."""
  from numpy import empty
  m = solib.psrad_get_kbdim()
  n = solib.psrad_get_klev()
  htngrt = empty([m,n], order="F", dtype="float64")
  htngrt_clr = empty([m,n], order="F", dtype="float64")
  n = n + 1
  flxd_lw_vr = empty([m,n], order="F", dtype="float64")
  flxd_lw_clr_vr = empty([m,n], order="F", dtype="float64")
  flxu_lw_vr = empty([m,n], order="F", dtype="float64")
  flxu_lw_clr_vr = empty([m,n], order="F", dtype="float64")
  C_htngrt = __get_double_p(htngrt)
  C_htngrt_clr = __get_double_p(htngrt_clr)
  C_flxd_lw_vr = __get_double_p(flxd_lw_vr)
  C_flxd_lw_clr_vr = __get_double_p(flxd_lw_clr_vr)
  C_flxu_lw_vr = __get_double_p(flxu_lw_vr)
  C_flxu_lw_clr_vr = __get_double_p(flxu_lw_clr_vr)
  solib.psrad_get_lw_fluxes(C_htngrt, C_htngrt_clr, C_flxd_lw_vr, C_flxd_lw_clr_vr, C_flxu_lw_vr, C_flxu_lw_clr_vr)
  return htngrt, htngrt_clr, flxd_lw_vr, flxd_lw_clr_vr, flxu_lw_vr, flxu_lw_clr_vr

def get_sw_fluxes():
  """Returns the following arrays with the results of the long wave routines:
* htngrt(float64[kbdim,klev]): Heating rate of total sky
* htngrt_clr(float64[kbdim,klev]): Heating rate of clear sky
For these variables, position [kbdim,klev-1] is the value at the surface layer and position [0] is at the TOA layer. NOTICE THIS IS THE OPPOSITE OF THE OTHER ROUTINE
* flxd_sw(float64[kbdim,klev+1]): Downward flux of total sky
* flxd_sw_clr(float64[kbdim,klev+1]): Downward flux of clear sky
* flxu_sw(float64[kbdim,klev+1]): Upward flux of total sky. 
* flxu_sw_clr(float64[kbdim,klev+1]): Upward flux of clear sky
For these variables, position [:,i] is the value at the top of layer [i], with 0 being at the TOA and [:,klev] being the surface.
* vis_frc_sfc(float64[kbdim]): Visible fraction of net surface radiation
* par_dn_sfc(float64[kbdim]): Photosynthetically active fraction of net surface radiation
* nir_dff_frc(float64[kbdim]): Diffuse near-infrared fraction
* vis_dff_frc(float64[kbdim]): Diffuse visible fraction
* par_dff_frc(float64[kbdim]): Diffuse photosynthetically active fraction"""
  from numpy import empty
  m = solib.psrad_get_kbdim()
  n = solib.psrad_get_klev()
  htngrt = empty([m,n], order="F", dtype="float64")
  htngrt_clr = empty([m,n], order="F", dtype="float64")
  n = n + 1
  flxd_sw = empty([m,n], order="F", dtype="float64")
  flxd_sw_clr = empty([m,n], order="F", dtype="float64")
  flxu_sw = empty([m,n], order="F", dtype="float64")
  flxu_sw_clr = empty([m,n], order="F", dtype="float64")
  vis_frc_sfc = empty(m, order="F", dtype="float64")
  par_dn_sfc = empty(m, order="F", dtype="float64")
  nir_dff_frc = empty(m, order="F", dtype="float64")
  vis_dff_frc = empty(m, order="F", dtype="float64")
  par_dff_frc = empty(m, order="F", dtype="float64")

  C_htngrt = __get_double_p(htngrt)
  C_htngrt_clr = __get_double_p(htngrt_clr)
  C_flxd_sw = __get_double_p(flxd_sw)
  C_flxd_sw_clr = __get_double_p(flxd_sw_clr)
  C_flxu_sw = __get_double_p(flxu_sw)
  C_flxu_sw_clr = __get_double_p(flxu_sw_clr)
  C_vis_frc_sfc = __get_double_p(vis_frc_sfc)
  C_par_dn_sfc = __get_double_p(par_dn_sfc)
  C_nir_dff_frc = __get_double_p(nir_dff_frc)
  C_vis_dff_frc = __get_double_p(vis_dff_frc)
  C_par_dff_frc = __get_double_p(par_dff_frc)

  solib.psrad_get_sw_fluxes(C_htngrt, C_htngrt_clr, C_flxd_sw, C_flxd_sw_clr, C_flxu_sw, C_flxu_sw_clr) 
  solib.psrad_get_sw_diags_old(C_vis_frc_sfc, C_par_dn_sfc, C_nir_dff_frc, C_vis_dff_frc, C_par_dff_frc)
  return htngrt, htngrt_clr, flxd_sw, flxd_sw_clr, flxu_sw, flxu_sw_clr, vis_frc_sfc, par_dn_sfc, nir_dff_frc, vis_dff_frc, par_dff_frc




# -*- coding: utf-8 -*-
import collections
import logging

import typhon
import netCDF4
import numpy as np
from scipy.interpolate import interp1d
from xarray import Dataset, DataArray

from konrad import constants
from konrad import utils
from konrad.convection import (Convection, HardAdjustment)
from konrad.humidity import (Humidity, FixedRH)
from konrad.lapserate import (LapseRate, MoistLapseRate)
from konrad.surface import (Surface, SurfaceHeatCapacity)
from konrad.upwelling import (Upwelling, NoUpwelling)


__all__ = [
    'Atmosphere',
]

logger = logging.getLogger(__name__)

atmosphere_variables = [
    'T',
    'H2O',
    'N2O',
    'O3',
    'CO2',
    'CO',
    'CH4',
]


class Atmosphere(Dataset):
    """Abstract base class to define requirements for atmosphere models."""
    def __init__(self, convection=None, humidity=None, surface=None,
                 lapse=None, upwelling=None, **kwargs):
        """Create an atmosphere model.

       Parameters:
             convection (konrad.humidity.Convection): Convection scheme.
                Defaults to ``konrad.convection.HardAdjustment``.
             humidity (konrad.humidity.Humidity): Humidity handler.
                Defaults to ``konrad.humidity.FixedRH``.
             surface (konrad.surface.Surface): Surface model.
                Defaults to ``konrad.surface.SurfaceHeatCapacity``.
             lapse (konrad.lapse.LapseRate): Lapse rate handler.
                Defaults to ``konrad.lapserate.MoistLapseRate``.
        """
        # Initialize ``xarray.Dataset`` with given positional args and kwargs.
        super().__init__(**kwargs)

        # Check input types.
        surface = utils.return_if_type(surface, 'surface',
                                       Surface, SurfaceHeatCapacity())

        humidity = utils.return_if_type(humidity, 'humidity',
                                        Humidity, FixedRH())

        convection = utils.return_if_type(convection, 'convection',
                                          Convection, HardAdjustment())

        lapse = utils.return_if_type(lapse, 'lapse',
                                     LapseRate, MoistLapseRate())

        upwelling = utils.return_if_type(upwelling, 'upwelling',
                                         Upwelling, NoUpwelling())

        # Set additional attributes for the Atmosphere object. They can be
        # accessed through point notation but do not need to be of type
        # ``xarrayy.DataArray``.
        self.attrs.update({
            'convection': convection,
            'humidity': humidity,
            'lapse': lapse,
            'surface': surface,
            'upwelling': upwelling,
        })

    def adjust(self, heatingrate, timestep, **kwargs):
        """Adjust temperature according to given heating rates.

        Parameters:
            heatingrate (ndarray): Radiative heatingrate [K/day].
            timestep (float): Timestep width [day].
        """
        # Caculate critical lapse rate.
        lapse = self.lapse.get(self)

        # Apply heatingrates to temperature profile.
        self['T'] += heatingrate * timestep

        # Convective adjustment
        self.convection.stabilize(atmosphere=self, lapse=lapse,
                                  timestep=timestep)

        # Upwelling induced cooling
        self.upwelling.cool(atmosphere=self, radheat=heatingrate[0, :],
                            timestep=timestep)

        # Calculate the geopotential height field.
        self.calculate_height()

        # Preserve the initial relative humidity profile.
        self['H2O'][0, :] = self.humidity.get(
            plev=self.get_values('plev'),
            T=self.get_values('T', keepdims=False),
            z=self.get_values('z', keepdims=False),
            p_tropo=self.get_convective_top(heatingrate[0, :]),
            T_surface=self.surface['temperature'].values[-1],
        )

    @classmethod
    def from_atm_fields_compact(cls, atmfield, **kwargs):
        """Convert an ARTS atm_fields_compact [0] into an atmosphere.

        [0] http://arts.mi.uni-hamburg.de/docserver-trunk/variables/atm_fields_compact

        Parameters:
            atmfield (typhon.arts.types.GriddedField4): A compact set of
                atmospheric fields.
        """
        # Create a Dataset with time and pressure dimension.
        plev = atmfield.grids[1]
        phlev = utils.phlev_from_plev(plev)
        d = cls(coords={'plev': plev,  # pressure level
                        'time': [0],  # time dimension
                        'phlev': phlev,  # pressure at halflevels
                        },
                **kwargs,
                )

        for var in atmosphere_variables:
            # Get ARTS variable name from variable description.
            arts_key = constants.variable_description[var].get('arts_name')

            # Extract profile from atm_fields_compact
            profile = atmfield.get(arts_key, keep_dims=False)
            d[var] = DataArray(profile[np.newaxis, :], dims=('time', 'plev',))

        # Calculate the geopotential height.
        d.calculate_height()

        # Append variable descriptions to the Dataset.
        utils.append_description(d)

        return d

    @classmethod
    def from_xml(cls, xmlfile, **kwargs):
        """Read atmosphere from XML file containing an ARTS atm_fields_compact.

        Parameters:
            xmlfile (str): Path to XML file.
        """
        # Read the content of given XML file.
        griddedfield = typhon.arts.xml.load(xmlfile)

        # Check if the XML file contains an atm_fields_compact (GriddedField4).
        arts_type = typhon.arts.utils.get_arts_typename(griddedfield)
        if arts_type != 'GriddedField4':
            raise TypeError(
                'XML file contains "{}". Expected "GriddedField4".'.format(
                    arts_type)
            )

        return cls.from_atm_fields_compact(griddedfield, **kwargs)

    @classmethod
    def from_dict(cls, dictionary, **kwargs):
        """Create an atmosphere model from dictionary values.

        Parameters:
            dictionary (dict): Dictionary containing ndarrays.
        """
        # TODO: Currently working for good-natured dictionaries.
        # Consider allowing a more flexibel user interface.

        # Create a Dataset with time and pressure dimension.
        plev = dictionary['plev']
        phlev = utils.phlev_from_plev(plev)
        d = cls(coords={'plev': plev,  # pressure level
                        'time': [0],  # time dimension
                        'phlev': phlev,  # pressure at halflevels
                        },
                **kwargs,
                )

        for var in atmosphere_variables:
            d[var] = DataArray(dictionary[var], dims=('time', 'plev',))

        # Calculate the geopotential height.
        d.calculate_height()

        # Append variable descriptions to the Dataset.
        utils.append_description(d)

        return d

    @classmethod
    def from_netcdf(cls, ncfile, timestep=-1, **kwargs):
        """Create an atmosphere model from a netCDF file.

        Parameters:
            ncfile (str): Path to netCDF file.
            timestep (int): Timestep to read (default is last timestep).
        """
        with netCDF4.Dataset(ncfile) as dataset:
            # Create a Dataset with time and pressure dimension.
            plev = dataset.variables['plev'][:]
            phlev = utils.phlev_from_plev(plev)
            d = cls(coords={'plev': plev,  # pressure level
                            'time': [0],  # time dimension
                            'phlev': phlev,  # pressure at halflevels
                            },
                    **kwargs,
                    )

            for var in atmosphere_variables:
                d[var] = DataArray(
                    data=dataset.variables[var][[timestep], :],
                    dims=('time', 'plev',)
                )

        # Calculate the geopotential height.
        d.calculate_height()

        # Append variable descriptions to the Dataset.
        utils.append_description(d)

        return d

    def to_atm_fields_compact(self):
        """Convert an atmosphere into an ARTS atm_fields_compact."""
        # Store all atmosphere variables including geopotential height.
        variables = atmosphere_variables + ['z']

        # Get ARTS variable name from variable description.
        species = [constants.variable_description[var].get('arts_name')
                   for var in variables]

        # Create a GriddedField4.
        atmfield = typhon.arts.types.GriddedField4()

        # Set grids and their names.
        atmfield.gridnames = ['Species', 'Pressure', 'Longitude', 'Latitude']
        atmfield.grids = [
            species, self['plev'].values, np.array([]), np.array([])
        ]

        # The profiles have to be passed in "stacked" form, as an ndarray of
        # dimensions [species, pressure, lat, lon].
        atmfield.data = np.vstack(
            [self[var].values.reshape(1, self['plev'].size, 1, 1)
             for var in variables]
        )
        atmfield.dataname = 'Data'

        # Perform a consistency check of the passed grids and data tensor.
        atmfield.check_dimension()

        return atmfield

    def to_netcdf(self, *args, **kwargs):
        """Write atmosphere contents to a netCDF file.

        Note:
             This is a wrapper simple for `xr.Dataset.to_netcdf`. The method
             converts all attributes (e.g. humidity, convection) to strings.
             This is necessary because the netCDF standard only allows
             string attributes.

        """
        def stringify(obj):
            """Return strings, integers and floats. Else return class name."""
            if isinstance(obj, (str, int, float)):
                return obj
            else:
                return obj.__class__.__name__

        attributes = self.attrs.copy()

        self.attrs = {key: stringify(val) for key, val in self.attrs.items()}

        super().to_netcdf(*args, **kwargs)

        self.attrs = attributes

    def refine_plev(self, pgrid, axis=1, **kwargs):
        """Refine the pressure grid of an atmosphere object.

        Note:
              This method returns a **new** object,
              the original object is maintained!

        Parameters:
              pgrid (ndarray): New pressure grid [Pa].
              axis (int): Index of pressure axis (should be 1).
                This keyword is only there for possible changes in future.
            **kwargs: Additional keyword arguments are collected
                and passed to :func:`scipy.interpolate.interp1d`

        Returns:
              Atmosphere: A **new** atmosphere object.
        """
        # Initialize an empty directory to fill it with interpolated data.
        # The dictionary is later used to create a new object using the
        # Atmosphere.from_dict() classmethod. This allows to circumvent the
        # fixed dimension size in xarray.DataArrays.
        datadict = dict()

        datadict['plev'] = pgrid  # Store new pressure grid.

        # Loop over all atmospheric variables...
        for variable in atmosphere_variables:
            # and create an interpolation function using the original data.
            f = interp1d(self['plev'].values, self[variable],
                         axis=axis, fill_value='extrapolate', **kwargs)

            # Store the interpolated new data in the data directory.
            datadict[variable] = DataArray(f(pgrid), dims=('time', 'plev'))

        # Create a new atmosphere object from the filled data directory.
        # This method also calculates the new phlev coordinates.
        new_atmosphere = type(self).from_dict(datadict)

        # Keep attributes of original atmosphere object.
        # This is **extremely** important because references to e.g. the
        # convection scheme or the humidity handling are stored as attributes!
        new_atmosphere.attrs = {**self.attrs}

        # Calculate the geopotential height.
        new_atmosphere.calculate_height()

        # Append variable descriptions to the Dataset.
        utils.append_description(new_atmosphere)

        return new_atmosphere

    # TODO: This function could handle the nasty time dimension in the future.
    # Allowing to set two-dimensional variables using a 1d-array, if one
    # coordinate has the dimension one.
    def set(self, variable, value):
        """Set the values of a variable.

        Parameters:
            variable (str): Variable key.
            value (float or ndarray): Value to assign to the variable.
                If a float is given, all values are filled with it.
        """
        if isinstance(value, collections.Container):
            self[variable].values[0, :] = value
        else:
            self[variable].values.fill(value)

    def get_values(self, variable, keepdims=True):
        """Get values of a given variable.

        Parameters:
            variable (str): Variable key.
            keepdims (bool): If this is set to False, single-dimensions are
                removed. Otherwise dimensions are keppt (default).

        Returns:
            ndarray: Array containing the values assigned to the variable.
        """
        if keepdims:
            return self[variable].values
        else:
            return self[variable].values.ravel()

    def calculate_height(self):
        """Calculate the geopotential height."""
        g = constants.earth_standard_gravity

        plev = self['plev'].values  # Air pressure at full-levels.
        phlev = self['phlev'].values  # Air pressure at half-levels.
        T = np.hstack((self.surface.temperature, self['T'][0, :]))
        p = np.hstack((phlev[0], plev))
        # Calculate the air density from current atmospheric state.
        rho = typhon.physics.density(p, T)

        dp = np.hstack((np.array([plev[0] - phlev[0]]), np.diff(plev)))
        rho_phlev = interp1d(p, rho)(phlev[:-1])
        # Use the hydrostatic equation to calculate geopotential height from
        # given pressure, density and gravity.
        z = np.cumsum(-dp / (rho_phlev * g))

        # If height is already in Dataset, update its values.
        if 'z' in self:
            self['z'].values[0, :] = z
        # Otherwise create the DataArray.
        else:
            self['z'] = DataArray(z[np.newaxis, :], dims=('time', 'plev'))

    def get_lapse_rates(self):
        """Calculate the temperature lapse rate at each level."""
        lapse_rate = np.diff(self['T'][0, :]) / np.diff(self['z'][0, :])
        lapse_rate = typhon.math.interpolate_halflevels(lapse_rate)
        lapse_rate = np.append(lapse_rate[0], lapse_rate)
        return np.append(lapse_rate, lapse_rate[-1])

    def get_potential_temperature(self, p0=1000e2):
        """Calculate the potential temperature.

        .. math::
            \theta = T \cdot \left(\frac{p_0}{P}\right)^\frac{2}{7}

        Parameters:
              p0 (float): Pressure at reference level [Pa].

        Returns:
              ndarray: Potential temperature [K].
        """
        # Get view on temperature and pressure arrays.
        T = self['T'].values[0, :]
        p = self['plev'].values

        # Calculate the potential temperature.
        return T * (p0 / p) ** (2 / 7)

    def get_static_stability(self):
        """Calculate the static stability.

        .. math::
            \sigma = - \frac{T}{\Theta} \frac{\partial\Theta}{\partial p}

        Returns:
              ndarray: Static stability [K/Pa].
        """
        # Get view on temperature and pressure arrays.
        t = self['T'].values[0, :]
        p = self['plev'].values

        # Calculate potential temperature and its vertical derivative.
        theta = self.get_potential_temperature()
        dtheta = np.diff(theta) / np.diff(p)

        return -(t / theta)[:-1] * dtheta

    def get_diabatic_subsidence(self, radiative_cooling):
        """Calculate the diabatic subsidence.

        Parameters:
              radiative_cooling (ndarray): Radiative cooling rates.
                Positive values for heating, negative values for cooling!

        Returns:
            ndarray: Diabatic subsidence [Pa/day].
        """
        sigma = self.get_static_stability()

        return -radiative_cooling[:-1] / sigma

    def get_subsidence_convergence_max(self, radiative_cooling, pmin=10e2):
        """Return index of maximum subsidence convergence.

        Parameters:
            radiative_cooling (ndarray): Radiative cooling rates.
                Positive values for heating, negative values for cooling!
            pmin (float): Lower pressure threshold. The cold point has to
                be below (higher pressure, lower height) that value.

        Returns:
              float: Pressure of maxixum subsidence divergence [Pa].
        """
        plev = self['plev'].values
        omega = self.get_diabatic_subsidence(radiative_cooling)
        domega = np.diff(omega) / np.diff(plev[:-1])

        # The found maximum is off by 1 due to numerical differentiation in
        # the subsidence calculation. Therefore, return the pressure above.
        max_index = np.argmax(domega[plev[:-2] > pmin]) + 1
        max_plev = plev[max_index]

        # Store found index and pressure level to atmosphere.
        self['diabatic_convergence_max_index'] = DataArray([max_index],
                                                           dims=('time',))
        self['diabatic_convergence_max_plev'] = DataArray([max_plev],
                                                          dims=('time',))

        return max_plev

    def get_convective_top(self, heatingrate, lim=-0.1):
        """Find the pressure where the radiative heating has a certain value.

        Note:
            In the HardAdjustment case, for a contop temperature that is not
            dependent on the number of distribution of pressure levels, it is
            better to take a value of lim not equal or very close to zero.

        Parameters:
            heatingrate (ndarray): Radiative heating rate [K/day].
            lim (float): Threshold value [K/day].

        Returns:
            float: Pressure at height of convective top [Pa].
        """
        p = self['plev'].values[:]
        T = self['T'].values[-1, :]

        # NOTE: `np.argmax` returns the first occurence of the maximum value.
        # In this example, the index of the first `True` value,
        # corresponding to the convective top, is returned.
        contop_i = int(np.argmax(heatingrate > lim))

        # Create auxiliary arrays storing the Qr, T and p values above and
        # below the threshold value. These arrays are used as input for the
        # interpolation in the next step.
        heat_array = np.array([heatingrate[contop_i-1], heatingrate[contop_i]])
        p_array = np.array([p[contop_i-1], p[contop_i]])
        T_array = np.array([T[contop_i-1], T[contop_i]])

        # Interpolate the pressure value where the heatingrate # equals `lim`.
        contop_plev = interp1d(heat_array, p_array)(lim)
        contop_T = interp1d(heat_array, T_array)(lim)

        # Store interpolated pressure level and temperature values.
        self['convective_top_plev'] = DataArray(
            [contop_plev], dims=('time',))
        self['convective_top_temperature'] = DataArray(
            [contop_T], dims=('time',))

        return contop_plev

    def get_coldpoint_plev(self, pmin=10e2):
        """Return the cold point pressure.

        Parameters:
            pmin (float): Minimum pressure threshold. The function does not
                return pressure values smaller than this. This prevents
                finding the upper most level, which is likely to be the
                coldest level.
        """
        T = self['T'].values[-1, :]
        plev = self['plev'].values[:]

        return plev[np.argmin(T[plev > pmin])]

    def tracegases_rcemip(self):
        """Set trace gas concentrations according to the RCE-MIP configuration.

        The volume mixing ratios are following the values for the
        RCE-MIP (Wing et al. 2017) and constant throughout the atmosphere.
        """
        concentrations = {
            'CO2': 348e-6,
            'CH4': 1650e-9,
            'N2O': 306e-9,
            'CO': 0,
        }
        plev = self['plev'].values

        # Assign constant VMR values to all atmosphere levels.
        for gas, vmr in concentrations.items():
            self[gas][0, :] = vmr * np.ones(plev.shape)

        # Determine the ozone profile according to RCEMIP.
        self['O3'][0, :] = utils.ozone_profile_rcemip(plev)



# -*- coding: utf-8 -*-
"""Implementation of a radiative-convective equilibrium model (RCE).
"""
import logging
from datetime import datetime

import numpy as np

from konrad import utils
from konrad.radiation import RRTMG


logger = logging.getLogger(__name__)

__all__ = [
    'RCE',
]


class RCE:
    """Interface to control the radiative-convective equilibrium simulation.

    Examples:
        Create an object to setup and run a simulation:
        >>> import konrad
        >>> rce = konrad.RCE(...)
        >>> rce.run()
    """
    def __init__(self, atmosphere, radiation=None, outfile=None, experiment='',
                 timestep=1, delta=0.01, writeevery=1, max_iterations=5000):
        """Set-up a radiative-convective model.

        Parameters:
            atmosphere (Atmosphere): `konrad.atmosphere.Atmosphere`.
            outfile (str): netCDF4 file to store output.
            experiment (str): Experiment description (stored in netCDF).
            timestep (float): Iteration time step in days.
            delta (float): Stop criterion. If the heating rate is below this
                threshold for all levels, skip further iterations.
            writeevery(int or float): Set frequency in which to write output.
                int: Every nth timestep is written.
                float: Every nth day is written.
            max_iterations (int): Maximum number of iterations.
        """
        # Sub-models.
        self.atmosphere = atmosphere
        if radiation is None:
            self.radiation = RRTMG()
        else:
            self.radiation = radiation

        # Control parameters.
        self.delta = delta
        self.timestep = timestep
        self.writeevery = writeevery
        self.max_iterations = max_iterations

        # TODO: Maybe delete? One could use the return value of the radiation
        # model directly.
        self.heatingrates = None

        # Internal variables.
        self.converged = False
        self.niter = 0

        self.outfile = outfile
        self.experiment = experiment

        logging.info('Created Konrad object:\n{}'.format(self))

    def __repr__(self):
        retstr = '{}(\n'.format(self.__class__.__name__)
        # Loop over all public object attributes.
        for a in filter(lambda k: not k.startswith('_'), self.__dict__):
            retstr += '    {}={},\n'.format(a, getattr(self, a))
        retstr += ')'

        return retstr

    def get_hours_passed(self):
        """Return the number of house passed since model start.

        Returns:
            float: Hours passed since model start.
        """
        return self.niter * 24 * self.timestep

    def calculate_heatingrates(self):
        """Use the radiation sub-model to calculate heatingrates."""
        self.heatingrates = self.radiation.get_heatingrates(
            atmosphere=self.atmosphere,
            )

    def is_converged(self):
        """Check if the atmosphere is in radiative-convective equilibrium.

        Returns:
            bool: ``True`` if converged, else ``False``.
        """
        return np.all(np.abs(self.atmosphere['deltaT']) < self.delta)

    def check_if_write(self):
        """Check if current timestep should be appended to output netCDF.

        Do not write, if no output file is specified.

        Returns:
            bool: True, if timestep should be written.
        """
        if self.outfile is None:
            return False

        if isinstance(self.writeevery, int):
            return self.niter % self.writeevery == 0
        elif isinstance(self.writeevery, float):
            # Add `0.5 * dt` to current timestep to make float comparison more
            # robust. Otherwise `3.3 % 3 < 0.3` is True.
            r = (((self.niter + 0.5) * self.timestep) % self.writeevery)
            return r < self.timestep
        else:
            raise TypeError('Only except input of type `float` or `int`.')

    # TODO: Consider implementing netCDF writing in a cleaner way. Currently
    # variables from different Datasets are hard to distinguish. Maybe
    # dive into the group mechanism in netCDF.
    def create_outfile(self):
        """Create netCDF4 file to store simulation results."""
        data = self.atmosphere.merge(self.heatingrates, overwrite_vars='H2O')
        data.merge(self.atmosphere.surface, inplace=True)

        # Add experiment and date information to newly created netCDF file.
        data.attrs.update(experiment=self.experiment)
        data.attrs.update(date=datetime.now().strftime("%Y-%m-%d %H:%M"))

        # Not all Radiation classes provide an `solar_constant` attribute.
        # For thos who do (e.g. `RRTMG`) store the value in the netCDF file.
        if hasattr(self.radiation, 'solar_constant'):
            data.attrs.update(solar_constant=self.radiation.solar_constant)

        # The `Atmosphere.to_netcdf()` function is overloaded and able to
        # handle attributes in a proper way (saving the object's class name).
        data.to_netcdf(self.outfile, mode='w', unlimited_dims=['time'])

        logger.info(f'Created "{self.outfile}".')

    def append_to_netcdf(self):
        """Append the current atmospheric state to the netCDF4 file specified
        in ``self.outfile``.
        """
        data = self.atmosphere.merge(self.heatingrates, overwrite_vars='H2O')
        data.merge(self.atmosphere.surface, inplace=True)

        utils.append_timestep_netcdf(
            filename=self.outfile,
            data=data,
            timestamp=self.get_hours_passed(),
            )

    def run(self):
        """Run the radiative-convective equilibrium model."""
        logger.info('Start RCE model run.')

        # Initialize surface pressure to be equal to lowest half-level
        # pressure. This is consistent with handling in PSrad.
        self.atmosphere.surface['pressure'] = self.atmosphere['phlev'][0]

        # Main loop to control all model iterations until maximum number is
        # reached or a given stop criterion is fulfilled.
        while self.niter < self.max_iterations:
            if self.niter % 100 == 0:
                # Write every 100th time step in loglevel INFO.
                logger.info(f'Enter iteration {self.niter}.')
            else:
                # All other iterations are only logged in DEBUG level.
                logger.debug(f'Enter iteration {self.niter}.')

            # Adjust the solar angle according to current time.
            self.radiation.adjust_solar_angle(self.get_hours_passed() / 24)

            # Caculate shortwave, longwave and net heatingrates.
            # Afterwards, they are accesible throug ``self.heatingrates``.
            self.calculate_heatingrates()

            # Apply heatingrates/fluxes to the the surface.
            self.atmosphere.surface.adjust(
                sw_down=self.heatingrates['sw_flxd'].values[0, 0],
                sw_up=self.heatingrates['sw_flxu'].values[0, 0],
                lw_down=self.heatingrates['lw_flxd'].values[0, 0],
                lw_up=self.heatingrates['lw_flxu'].values[0, 0],
                timestep=self.timestep,
            )

            # Save the old temperature profile. They are compared with
            # adjusted values to check if the model has converged.
            T = self.atmosphere['T'].values.copy()

            # Apply heatingrates to the temperature profile.
            self.atmosphere.adjust(
                self.heatingrates['net_htngrt'],
                self.timestep,
                surface=self.atmosphere.surface,
                )

            # Calculate temperature change for convergence check.
            self.atmosphere['deltaT'] = self.atmosphere['T'] - T

            # Check, if the current iteration is scheduled to be written.
            if self.check_if_write():
                # If we are in the first iteration, a new is created...
                if self.niter == 0:
                    self.create_outfile()
                # ... otherwise we just append.
                else:
                    self.append_to_netcdf()

            # Check if the model run has converged to an equilibrium state.
            if self.is_converged():
                # If the model is converged, skip further iterations. Success!
                logger.info(f'Converged after {self.niter} iterations.')
                break
            # Otherweise increase the iteration count and go on.
            else:
                self.niter += 1
        else:
            logger.info('Stopped after maximum number of iterations.')
